/* $Id$

This file is part of libmspgl
Copyright © 2007  Mikko Rasa, Mikkosoft Productions
Distributed under the LGPL
*/

#ifndef MSP_GL_OBJECT_H_
#define MSP_GL_OBJECT_H_

#include <vector>
#include <msp/datafile/collection.h>

namespace Msp {
namespace GL {

class Material;
class Mesh;
class ObjectInstance;
class Program;
class ProgramData;
class Texture;

/**
Stores data for a complete 3D object.  This includes a mesh, a shader and data
for it and a number of textures.  Only the mesh is mandatory, other components
are optional.

See also class ObjectInstance.
*/
class Object
{
private:
	Mesh *mesh;
	std::vector<Texture *> textures;
	Program *shprog;
	ProgramData *shdata;
	Material *material;

public:
	class Loader: public DataFile::Loader
	{
	public:
		typedef DataFile::Collection Collection;

	protected:
		Object &obj;
		Collection &coll;
	private:
		std::vector<std::string> textures;
	
	public:
		Loader(Object &, Collection &);
		~Loader();

		Object &get_object() const { return obj; }
		Collection &get_collection() const { return coll; }
	private:
		void shader(const std::string &);
		void texture(const std::string &);
	};

	Object();
	~Object();

	Program *get_shader() const { return shprog; }

	/**
	Renders the object.  If an ObjectInstance is provided, its hook functions
	are called.
	*/
	void render(const ObjectInstance * =0) const;

	/**
	Renders multiple instances of the object in one go.  This may be a
	performance improvement, as the object-specific render setup only has to be
	done once.
	*/
	void render(const std::list<const ObjectInstance *> &) const;
private:
	void setup_render() const;
	void finish_render() const;
};

} // namespace GL
} // namespace Msp

#endif
